#include <iostream> 
#include <fstream> 
#include <string>
#include <vector>
#include <deque> 
#include <sstream>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <math.h>
#include <stdlib.h>
#include "ProcSched.h"
#include "nsSysteme.h"

using namespace nsSysteme;

using namespace std;

namespace ProcSched {

  int bRand(int upLimit) {
     return (1 + static_cast<int>(static_cast<double>(upLimit)*rand()/(RAND_MAX+1.0)));
  }



  string       ProcInfo::ProcData::linePad;
  unsigned int ProcInfo::ProcData::reqLength;

  ProcInfo :: 
  ProcData :: 
  ProcData(const string &name) : progName           (name), 
				 procStatus         (STAT_WAITING), 
				 programCounter     (0),
				 currentInstructionRemainingTime(-1),
				 totalTime          (0),
				 totalReqMinDuration(0), 
				 totalReqCPUDuration(0), 
				 totalReqIODuration (0), 
				 totalReqSYSDuration(0), 
				 totalMinDuration   (0),
				 totalCPUDuration   (0), 
				 totalIODuration    (0), 
				 totalSYSDuration   (0),
				 CPUusage           (0),
				 waitDuration       (0),
                                 lastTimeEnQueued   (0),
				 priorityLevel      (0),
				 line2print         (0) {
    time_t t;
    time(&t);
    startTime  = t;
  }

  int ProcInfo::ProcData::parseProg(const string &fileLine, 
				    unsigned int lineNumber,
				    const string &fileName) {
    // syntaxe de chaque ligne : {compute,io,fork} {exact,approx,random} <nombre>
    // exact  == l'instruction "dure" exactement nombre secs  
    // approx == nombre sec + ou - x%, avec x<=20
    // random == entre 1 et nombre sec
    string instrT;
    ProcInstructionType t;
    ProcInstructionDurationType d(EXACT_TIME);
    istringstream buffLine(fileLine);
    if(buffLine >> instrT) {
      string instrK (instrT);
      for(unsigned int k = 0 ; k < instrK . size(); ++k) instrK[k] = tolower(instrK[k]);
      string instrD;
      if(buffLine >> instrD) {
	int r(0),v(0);
	if(buffLine >> r) {
	  string instrR (instrD);
	  for(unsigned int k = 0 ; k < instrR . size(); ++k) instrR[k] = tolower(instrR[k]);
	  if(instrR == "exact") {
	    d = EXACT_TIME;
	    v = r;
	  }
	  else if(instrR == "approx") {
	    d = APPROX_TIME;
	    v = bRand(r/5) * (bRand(50)>25?1:-1) + r; 
	  } 
	  else if(instrR == "random") {
	    d = RANDOM_TIME;
	    v = bRand(r);
	  }
	  else {
	    cerr << "\nERROR Unknown duration type " << instrR 
		 << " line " << lineNumber <<  " file " << fileName << ", skipping this program\n";
	    return(1);
	  }
	  if(instrK == "compute") {
	    t = DO_CPU;
	    totalReqCPUDuration += d;
	    totalCPUDuration    += v;
	  }
	  else if(instrK == "io") {
	    t = DO_IO;
	    totalReqIODuration += d;
	    totalIODuration    += v;
	  } 
	  else if(instrK == "fork") {
	    t = DO_FORK;
	    totalReqSYSDuration += d;
	    totalSYSDuration    += v;
	  }
	  else {
	    cerr << "\nERROR Unknown instruction " << instrT 
		 << " line " << lineNumber <<  " file " << fileName << ", skipping this program\n";
	    return(1);
	  }
	  proGram . push_back(ProcInstruction(t,d,r,v));
	  totalReqMinDuration += r;
	  totalMinDuration += v;
	  return (0);
	}
	else {
	  cerr << "\nERROR Expecting duration value as a third argument in line " 
	       << lineNumber <<  " file " << fileName << ", skipping this program\n";
	  return(1);
	}
      }
      else {
	cerr << "\nERROR Expecting duration type as a second argument in line " 
	     << lineNumber <<  " file " << fileName << ", skipping this program\n";
	return(1);
      }
    }
    else {
      cerr << "\nERROR Expecting instruction type as a first argument in line " 
	   << lineNumber <<  " file " << fileName << ", skipping this program\n";
      return(1);
    }
    return(2);
  }
    
  ProcInfo::ProcInfo(const string &fileList): averageWaitDuration(0), progPerLine(7) {	
    // on ouvre et lit ligne par ligne les fichiers passes en arguments, un par un
    // la methode ProcData::parseProg() s'occupe du parsing effectif de chaque ligne
    // elle rend zero si et seulement si tout va bien, i.e. pas d'erreur de syntaxe 
    istringstream buffStr(fileList);
    longestProgramLength = 0;
    for(string fileName; buffStr >> fileName;) { // pour chaque fichier
      ifstream progFile (fileName . c_str());
      if(!progFile) {
	cerr << "ERROR Bad file name " << fileName << ", skipping.\n";
	continue;
      }
      cerr << "INFO Reading " << fileName << ", lines " ;
      procData . push_back(ProcData(fileName));
      bool qOk (true);
      unsigned int lineNumber (0);
      for(string fileLine ;  // pour chaque ligne du fichier courant
	  qOk && getline(progFile, fileLine) ; ++lineNumber) {
	cerr << lineNumber << " ";
	if(procData . back() . parseProg(fileLine,lineNumber,fileName)) { // oops
	  procData . pop_back();  // ce fichier comporte au moins une erreur de syntaxe
	  qOk = false;            // et alors on y renonce
	}
      } // fin de la boucle pour chaque ligne du fichier courant
      if(qOk) { // tout va bien pour ce fichier, quelques initialisations restent a faire
	cerr << "ok.\n";
	procData . back() . currentInstructionRemainingTime = procData . back() . proGram[0] . instructionDecidedDuration;
	if(longestProgramLength < procData . back() . proGram . size()) {
	  longestProgramLength = procData . back() . proGram . size();
	}
      }    
    } // fin de la boucle pour chaque fichier 
    outstandingProcCount = procData . size();
    if(outstandingProcCount) {
      procData[0] . setLinePad("            ");
    } 
    for(unsigned int k = 0 ; k < progPerLine * (ProcData::reqLength + 3); ++k) {
      lineSeparator += '-';
    }

  } // fin de ProcInfo::ProcInfo()

  // **************************************************** //
  // Ce qui suit sert seulement pour le PRETTY PRINTING * //
  // **************************************************** //

  // L'idee de l'affichage est de le construire ligne par ligne
  // on veut des "rectangles" pour chaque processus, entoures
  // eventuellement de '+' ou '*' ou '|' selon l'etat (io, execution, fin)
  // on se fixe un nombre max de processus par rangee horizontale
  // (on va dire sept, car leur largeur est plus ou moins connue)
  // et on balaie les processus par groupes de sept
  // ProcInfo::displayProcInfo() est la fonction principale :
  // pour chaque ligne
  //    pour chaque processus
  //       afficher la ligne avec ProcData::nextLine2Str() laquelle
  //                 incremente ProcData::line2print et
  //                 appelle ProcData::progLine2Str()
  // 
  // on veut de plus souligner de '^^^' la ligne en cours d'execution
  // tout en gardant la structure (i.e. une ligne affichee par appel de
  // nextLine2Str()). on utilise alors les valeurs negatives comme bascule:
  // 
  // ProcData::nextLine2Str()
  //    si line2print negatif alors
  //        on affiche les '^^^' et on le rend positif et incremente
  //    sinon
  //        on appelle progLine2Str()
  //        si line2print == programCounter alors
  //            on rend line2print negatif
  //        sinon on l'incremente
  //        
 


  void ProcInfo::ProcData::setLinePad(const string &lp) {
    linePad = lp;
    reqLength = linePad . size();
  }

  string ProcInfo::ProcData::procStatus2Str(char *pad) const {
    switch(procStatus) {
    case  STAT_RUNNING: {
      if(pad) *pad = '*';
      return "running";
      break;
    }
    case  STAT_IO: {
      if(pad) *pad = '+';
      return "io";
      break;
    }
    case  STAT_SYS: {
      if(pad) *pad = '#';
      return "sys";
      break;
    }
    case  STAT_WAITING: {
      if(pad) *pad = ' ';
      return "wait";
      break;
    }
    case  STAT_TERMINATED: {
      if(pad) *pad = '|';
      return "terminated";
      break;
    }
    default: {
      cerr << "ERROR Unknown ProcStatus " << procStatus << "\n";
      if(pad) *pad = '?';
      return "unknownState";
    }
    }
  }

  string ProcInfo::ProcData::progLine2Str(const unsigned int lineNumber) const {
    if(lineNumber  >= proGram . size()) {
      return linePad;
    }
    ostringstream buffStr;
    string lineStr;
    switch(proGram[lineNumber] . instructionType) {
    case DO_CPU: lineStr = "cpu "; break;
    case DO_IO:  lineStr = "io  "; break;
    case DO_FORK:lineStr = "sys "; break;
    default: return"";
    }
    switch(proGram[lineNumber] . instructionDurationType) {
    case EXACT_TIME:  lineStr += "x "; break;
    case APPROX_TIME: lineStr += "a "; break;
    case RANDOM_TIME: lineStr += "r "; break;
    default: return"";
    }
    buffStr << proGram[lineNumber] . instructionRequestedDuration << " "
	    << proGram[lineNumber] . instructionDecidedDuration ;
    lineStr += buffStr . str();
    if(lineStr . size() < reqLength) {
      lineStr += linePad . substr(0,reqLength - lineStr . size());
    }
    return(lineStr);
  }

  // les fonctions suivantes font usage de linePad qui est une suite 
  // d'espaces blancs, en recuperant autant qu'il faut pour aboutir 
  // a une longueur fixe : on construit ainsi chaque element de 
  // ligne de processus de maniere reguliere

  //  . ligne proc .
  //  . ligne proc .
  //  . ... 

  string ProcInfo::ProcData::progCrtLine2Str() const {
    string lineStr;
    if(procStatus == STAT_TERMINATED) {
      lineStr = " ** END ** ";
      if(lineStr . size() < reqLength) {
	lineStr += linePad . substr(0,reqLength - lineStr . size());
      }
      return lineStr;
    }
    ostringstream buffstr;
    if(programCounter > proGram . size()) {
      buffstr << "?? crtLine# " << programCounter<< " outOfBound";
      lineStr = buffstr . str();
      if(lineStr . size() < reqLength) {
	lineStr += linePad . substr(0,reqLength - lineStr . size());
      }
      return lineStr;
    }
    const ProcInstruction &instr (proGram[programCounter]);
    const unsigned int progressBarLength = 
      static_cast<unsigned int>(ceil((reqLength - 1) * 
				     static_cast<double>(instr . instructionDecidedDuration - 
				      currentInstructionRemainingTime) / 
				     instr . instructionDecidedDuration));
    lineStr = "*";
    for(unsigned k= 0 ; k < progressBarLength ; ++k) {
      lineStr += "^";
    }
    if(lineStr . size() < reqLength) {
      lineStr += linePad . substr(0,reqLength - lineStr . size());
    }
    return lineStr;
  }    

  void ProcInfo::ProcData::resetLine2StrGen() {
    line2print = 0;
  }
  
  string ProcInfo::ProcData::nextLine2Str() {
    char c[2];
    c[1]='\0';
    procStatus2Str(c);
    if(line2print >= static_cast<int>(proGram . size())) {
      return(string(c) + linePad + c);
    }
    string lineStr;
    if(line2print < 0) { // i.e.on vient d'afficher ligne courante
      // lors de l'appel precedent, donc maintenant on affiche 
      // l'avancement de son execution
      lineStr = string(c) + progCrtLine2Str() + c;
      line2print = -(line2print + 1);
      ++line2print;
    }
    else {
      lineStr = string(c) + progLine2Str(line2print) + c;
      if(line2print == static_cast<int>(programCounter)) {
	line2print = -(line2print + 1);
      }
      else {
	++line2print;
      } 
    }
    return lineStr;
  }

  // fonction principale qui fait le balayage 
  

  //  . ligne1 proc1 .   ->   . ligne1 proc2 .   -> . ligne1 proc3 .  ... fin de ligne
  //
  //  et puis
  //
  //  . ligne2 proc1 .   ->   . ligne2 proc2 .   -> . ligne2 proc3 .  ... fin de ligne
  // 
  //  et ainsi de suite

  // cette fonction affiche autant de telles rangees que necessaire
  // (le nombre de processus pouvant croitre durant leur simulation <= 'fork') 

  void ProcInfo::displayProcInfo(ostream *s) {
    int proc2Print (procData . size());
    for(;proc2Print > 0; proc2Print -= progPerLine) {
      const unsigned int thisLineProcCount   = (proc2Print < progPerLine) ? proc2Print : progPerLine;
      const unsigned int firstProcOnThisLine = procData . size() - proc2Print;
      const unsigned int lastProcOnThisLine  = procData . size() - proc2Print + thisLineProcCount - 1;

      // d'abord les en-tetes de chaque rectangle
      for(unsigned int kProc = firstProcOnThisLine ; kProc <= lastProcOnThisLine ; ++kProc) {
	char c[2];
	c[1] = '\0';
	procData[kProc] . procStatus2Str(c);
	string lineStr;
	lineStr =  string(" ") + procData[kProc] . progName ;
	procData[kProc] . resetLine2StrGen();
	if(lineStr . size() < ProcData::reqLength) {
	  lineStr += ProcData::linePad . substr(0,ProcData::reqLength - lineStr . size());
	}
	else {
	  lineStr . resize(ProcData::reqLength);
	}
	(*s) << c << lineStr << c << " ";
      }
      (*s) << "\n";
      // maintenant les lignes
      for(unsigned int j = 0; j <= longestProgramLength ; ++j) {
	for(unsigned int kProc = firstProcOnThisLine ; kProc <= lastProcOnThisLine ; ++kProc) {
	  (*s) << procData[kProc] . nextLine2Str() << " " ;
	}
	(*s) << "\n";
      }
      // et maintenant les pieds de chaque rectangle
      for(unsigned int kProc = firstProcOnThisLine ; kProc <= lastProcOnThisLine ; ++kProc) {
	char c;
	string lineStr (procData[kProc] . procStatus2Str(&c));
	if(procData[kProc] . procStatus != STAT_TERMINATED) {
	  ostringstream buffstr;
	  buffstr  << " " << lineStr << " " << procData[kProc] . priorityLevel;
	  lineStr = buffstr . str();
	}
	if(lineStr . size() < ProcData::reqLength) {
	  lineStr += ProcData::linePad . substr(0,ProcData::reqLength - lineStr . size());
	}
	(*s) << c << lineStr << c << " " ;
      }
      (*s) << "\n";
      // on a enfin termine avec cette rangee de processus
    } // fin de la grande boucle proc/ligne 
    (*s) << lineSeparator << "\n";
  } 

  string ProcInfo::getProcName(const int procPid) const {
    return(procData[procPid] . progName);
  }


  // ***************************************** //
  // Ce qui suit sert pour les statistiques  * //
  // ***************************************** //
 
  void ProcInfo::dumpProcInfoStat(ostream *s) const {
    (*s)   << "#\tTt\tWt\tMin\tCPU\tIO\tSYS\n"; 
    for(unsigned int kProc = 0 ; kProc < procData . size() ; ++kProc) {
      (*s) << kProc << "\t" 
	   << procData[kProc] . totalTime           << "\t"
	   << procData[kProc] . waitDuration        << "\t"
	   << procData[kProc] . totalMinDuration    << "\t"
	   << procData[kProc] . totalCPUDuration    << "\t" 
	   << procData[kProc] . totalIODuration     << "\t" 
	   << procData[kProc] . totalSYSDuration    << "\n";
    }
  }

  void ProcInfo::ProcData::computeDuration(){
    // on utilise l'horloge systeme en faisant 
    // la difference entre les moments
    // l'horloge systeme rend le nombre de secondes
    // depuis l'Epoque
    time_t t;
    time(&t);
    totalTime = t - startTime;
  }

  void ProcInfo::ProcData::resetCPUusage(){
    CPUusage = 0;
  }

  void ProcInfo::ProcData::computeWaitDuration(){
    // on utilise l'horloge systeme en rajoutant
    // les differences successives entre les moments 
    // d'entree et de sortie dans la file d'attente
    time_t t;
    time(&t);
    waitDuration += t - lastTimeEnQueued; 
    lastTimeEnQueued = 0;
  }

  void ProcInfo::computeAverageWait(){
    averageWaitDuration = 0;
    for(unsigned int kProc = 0;
	kProc < procData . size(); ++kProc) {
      averageWaitDuration += procData[kProc] . waitDuration;
    }
    if(procData . size()) {
      averageWaitDuration /= procData . size();
    }
  }

  // *************************************************** //
  // Ce qui suit sert pour l'"execution" des processus * //
  // *************************************************** //


  // c'est en fait pour aider au scheduling avec le multilevel

  int ProcInfo::ProcData::computePriorityLevel(){
    priorityLevel = static_cast<int>((static_cast<double>(priorityLevel)) / 2.0L + 
      (static_cast<double>(CPUusage)) / 2.0L);
    resetCPUusage();
    return priorityLevel;
  }

  void ProcInfo::computePriorityLevel(){
    for(unsigned int kProc = 0;
	kProc < procData . size(); ++kProc) {
      if(procData[kProc] . procStatus != ProcInfo::STAT_TERMINATED) {
	procData[kProc] . computePriorityLevel();
      }
    }
  }

  // le modele d'execution choisi simule la preemption en allouant des
  // tranches de temps predefinies. ainsi on ne fait de fork() reel 
  // (i.e. Unix) que pour le IO, pour beneficier d'une horloge par
  // operation d'entree-sortie. en gros on dit au "moteur d'avancement"
  // 'tu as allotedTime secondes, fais avancer le processus choisi 
  // avec l'instruction en cours pendant autant de temps, 
  // et apres reviens'

  // cette tranche de temps s'applique seulement pour 'compute'
  // (on simplifie le modele et on considere que le 'fork'
  // prend un temps "borne" et "petit"). le io est de toutes manieres a part

  // de plus, pour assurer une plus grande flexibilite, on prevoit le cas ou 
  // la tranche ainsi allouee est PLUS GRANDE que ce qu'il y a a faire
  // dans l'instruction courante 'compute'. une petite boucle while()
  // avance pas-a-pas jusqu'a epuiser allotedTime ou recontrer un 'io' ou 
  // bien un 'fork' (autre simplification : le prix pour ne pas compter 
  // le temps du fork dans allotedTime est la renonciation au residu eventuel
  // si le fork() est rencontre dans ce while(): i.e. on sort completement
  // et on considere la tranche epuisee des qu'on a autre chose a faire que
  // du 'compute'

  // l'execution se fait a travers l'appel a ProcInfo::updateProcData()
  // avec l'ordre 'ADVANCE_PROC' (voir plus bas)

  // cette fonction delegue l'IO et le SYS (ici le fork) a deux fonctions
  // auxiliaires: ProcInfo::doIOIfCrtInstruction() ('says so') et respectivement
  // ProcInfo::doForkIfCrtInstruction(), lesquelles regardent d'abord 
  // si l'instruction en cours est une instruction de type io ou respectivement fork
  // sinon, elles sortent tout de suite et rendent zero
  // si oui, elle la traitent et rendent un nombre non-nul
  
  int ProcInfo::doIOIfCrtInstruction(const int procPid) {
    if(procData[procPid] . proGram[procData[procPid] . programCounter] . instructionType == DO_IO) {
     try {
      if(pid_t supportPid = Fork()) {
	procSupp2procData[supportPid] = procPid;
	procData[procPid] . procStatus = STAT_IO;
	procData[procPid] . currentInstructionRemainingTime = 0;
	return 1;
      }
      else {
	sleep(procData[procPid] . proGram[procData[procPid] . programCounter] . instructionDecidedDuration);
	exit(0);
      }
     }
     catch (const CExc & E) {
        cerr << "Probleme avec le fork() reel d'Unix: " << E << "\n";
     }
    }
    return 0;
  }
  
  int ProcInfo::doForkIfCrtInstruction(const int procPid) {
    if(procData[procPid] . proGram[procData[procPid] . programCounter] . instructionType == DO_FORK) {
      procData[procPid] . currentInstructionRemainingTime = 0;
      // d'abord le fils qui vient d'etre cree par "dedoublement"
      procData   . push_back(procData[procPid]);
      procData   . back() . procStatus = STAT_WAITING; 
      procData   . back() . waitDuration = 0;
      time_t t; time(&t);
      procData   . back() . startTime = t;
      scheduler -> enQueueProc(procData . size() - 1);
        ++outstandingProcCount;
      // et maintenant on prend soin du pere aussi
      procData[procPid]   . procStatus = STAT_SYS;
      displayProcInfo(&cerr);
      sleep(procData[procPid] . 
	    proGram[procData[procPid] . programCounter] . 
	    instructionDecidedDuration); // le pere se "repose" un peu 
      // et maintenant on avance le programCounter des deux  
      if(procData[procPid] . programCounter < procData[procPid] . proGram . size() - 1) {
	procData[procPid]  . programCounter ++ ;
	const int d = procData[procPid] . 
	  proGram[procData[procPid] . programCounter] . instructionDecidedDuration;
	procData . back()  . programCounter ++ ;
	procData . back()  . currentInstructionRemainingTime = d;
	procData[procPid]  . currentInstructionRemainingTime = d;
	procData[procPid]  . procStatus =  STAT_WAITING;
	scheduler -> enQueueProc(procPid);
      }
      else {
	procData . back()  . procStatus         =  STAT_TERMINATED;
	procData . back()  . computeDuration();
	procData[procPid]  . procStatus         = STAT_TERMINATED;
	procData[procPid]  . computeDuration();
	--outstandingProcCount;	--outstandingProcCount;	
      }
      return 1;
    }
    return 0;
  }
  
  // la fonction principale pour l'avancement instruction par instruction, ainsi que 
  // pour la prise en compte de la fin d'instructions de io ou compute
  // la fin de io est appelee depuis le gestionnaire de SIGCHLD 


  // pour ADVANCE_PROC, ProcInfo::updateProcData() rend le temps a comptabiliser dans
  // l'usage total de la CPU pour l'utilisateur, mais c'est elle qui fait le sleep()
  // pour simuler le travail effectif
  
  int ProcInfo::updateProcData(const int        procPid, 
			       const ProcInfo::ProcInfoOperType p, 
			       int              allotedTime/* = 0*/,
			       const pid_t      supportPid /* = 0*/) {

    if(p != IO_TERM && procPid == ProcInfo::invalidProcPid) {
      return 0;
    }
    switch(p) {
    case IO_TERM: {
      int thisProcPid = procSupp2procData[supportPid];
      if(procData[thisProcPid]  . programCounter < procData[thisProcPid] . proGram . size() - 1) {
	procData[thisProcPid]   . procStatus = STAT_WAITING; 
	scheduler              -> enQueueProc(thisProcPid);
	procData[thisProcPid]   . programCounter++;
	procData[thisProcPid]   . currentInstructionRemainingTime = 
	  procData[thisProcPid] . proGram[procData[thisProcPid] . programCounter] . instructionDecidedDuration;
      }
      else {
	procData[thisProcPid] . procStatus                      = STAT_TERMINATED;
	procData[thisProcPid] . computeDuration();
	--outstandingProcCount;	
      }
      displayProcInfo(&cerr);
      return 0;
    }
    case CPU_TERM: {
      if(procData[procPid] . procStatus != STAT_TERMINATED) {
      	procData[procPid] . procStatus = STAT_WAITING; 
	scheduler -> enQueueProc(procPid);
      }
      displayProcInfo(&cerr);
      return 0;
    }
    case ADVANCE_PROC: {
      const int r = doForkIfCrtInstruction(procPid);
      if(r) {
	return 0;
      }
      if(doIOIfCrtInstruction(procPid)) {
	return 0;
      }
      // ok, c'est du compute, on s'en occupe maintenant
      procData[procPid] . procStatus = STAT_RUNNING;
      int workTime (0);
      // cette boucle sert dans le cas ou le allotedTime est "assez" grand 
      while(allotedTime >= procData[procPid] . currentInstructionRemainingTime && 
	    procData[procPid] . programCounter < procData[procPid] . proGram . size() - 1) {

	procData[procPid] . CPUusage += procData[procPid] . currentInstructionRemainingTime;
	if(procData[procPid] . currentInstructionRemainingTime) { 
	  sleep(procData[procPid] . currentInstructionRemainingTime);
	  workTime += procData[procPid] . currentInstructionRemainingTime;
	}
	allotedTime -= procData[procPid] . currentInstructionRemainingTime;
	procData[procPid] . programCounter++;
	procData[procPid] . currentInstructionRemainingTime = procData[procPid] . proGram[procData[procPid] . programCounter] . instructionDecidedDuration;
	if(allotedTime > 0) { 
	  const int rr = doForkIfCrtInstruction(procPid);
	  if(rr) {
	    return 0;
	  }
	  if(doIOIfCrtInstruction(procPid)) {
	    return 0;
	  }
	}
      }
      // bon, on est au bout du programme ou de la tranche de temps
      if(allotedTime >= procData[procPid] . currentInstructionRemainingTime) {
	procData[procPid] . CPUusage += procData[procPid] . currentInstructionRemainingTime;
	if(procData[procPid] . currentInstructionRemainingTime) {
	  sleep(procData[procPid] . currentInstructionRemainingTime);
	  workTime += procData[procPid] . currentInstructionRemainingTime;
	}
	allotedTime -= procData[procPid] . currentInstructionRemainingTime;
	procData[procPid] . currentInstructionRemainingTime = 0;
	procData[procPid] . procStatus                      = STAT_TERMINATED;
	procData[procPid] . computeDuration();
	--outstandingProcCount;
      }
      else { // ah, c'est que la tranche de temps est presque epuisee
	// et en tout cas, plus petite que ce qui est necessaire pour finir 
	// TOUT ce qui reste de l'instruction en cours
	procData[procPid] . CPUusage += allotedTime;
	if(allotedTime) {
	  sleep(allotedTime);
	  workTime += allotedTime;
	}
	procData[procPid] . currentInstructionRemainingTime -= allotedTime;
	procData[procPid] . procStatus                      = STAT_RUNNING;
      }
      return workTime;
      break;
    }
    default: cerr << "ERROR Unknown ProcInfoOperType " << p << "\n";
      return -1;
      break;
    }
    return -2;
  }

  // pour simplifier la manipulation depuis le gestionnaire du signal

  ProcInfo * procInfo;

  
  // le gestionnaire du signal, pour prendre en compte la fin de IO

  void doIOTermAck(int sig) { 
    if(sig != SIGCHLD) return;
    int status;
    pid_t supportPid;
    try {
      while((supportPid = Waitpid(-1,&status,WNOHANG)) > 0) {
        procInfo -> updateProcData(-1, ProcInfo::IO_TERM, -1, supportPid);
        // dans cette fonction on "devine" de quel processus il s'agit
        // effectivement, a l'aide de la map<pid_t,int> faisant correspondre
        // les pid Unix reels aux pid des processus simules ayant lance des
        // operations de IO 
       }
    }
    catch(...) {}
  }

  // et maintenant la partie ordonnanceur

  // cette fonction reinitialise la donnee-membre lastTimeEnQueued 
  // chaque fois, pour permettre ensuite le calcul du temps d'attente
  // par additions successives des differences temporelles

  // tempsAttente = Somme(moment_sortie_i - moment_entree_i)

  // elle peut exceptionnellement ne pas le faire, et ceci est necessaire
  // pour le multilevel queue, lors du recalcul des priorites et donc des
  // compositions des files, car la on est au "milieu" d'intervalle d'attente, 
  // il y a juste un rearrangement, sans election effective


  
  Scheduler::SchedPolicy Scheduler::str2Policy(const string &p) {
    if(p == "rRobin")  return ROUND_ROBIN_SCHED;
    if(p == "fifo")    return FIFO_SCHED;
    if(p == "lifo")    return LIFO_SCHED;
    if(p == "mLevel")  return MULTILEVEL_FEEDBACK_SCHED;
    return INVALID_SCHED;
  }

  string Scheduler::policy2str() {
    switch(thePolicy) {
    case FIFO_SCHED: return "fifo";
    case LIFO_SCHED: return "lifo";
    case ROUND_ROBIN_SCHED: return "rRobin";
    case MULTILEVEL_FEEDBACK_SCHED: return "mLevel";
    default: return "Unknown policy in policy2str()";
    }
  }	


  void Scheduler::enQueueProc(const int procPid, const bool qSetTime /* = true */) {
    if(qSetTime) {
      time_t t;
      time(&t);
      pInfo -> procData[procPid] . lastTimeEnQueued = t;
    }
    switch(thePolicy) {
    case LIFO_SCHED: 
    case ROUND_ROBIN_SCHED: { 
      // TRAVAIL FAIT : - - - -
      // empilement procPid dans waitQueue par derriere
      waitQueue . push_back(procPid); 
      // - - - - - - - - - - - 
      break;
    }
    default:{
      cerr << "ERROR Unknown policy " << thePolicy << " in enQueueProc().\n";
      break;
    }
    }
  }

  // simple appel de la fonction precedente
  
  void Scheduler::enQueueAllProc(const bool qSetTime/* = true */) {
    for(unsigned int kProc = 0; kProc < pInfo -> procData . size(); ++kProc) {
      if(pInfo -> procData [kProc] . procStatus == ProcInfo::STAT_WAITING) {
	enQueueProc(kProc, qSetTime);
      }
    }
  }

  void Scheduler::displayQueue(ostream *s) {
    (*s) << "(" ;
    switch(thePolicy) {
    case FIFO_SCHED: {
      for(unsigned int kProc = 0; kProc < pInfo -> procData . size(); ++kProc) {
	if(pInfo -> procData[kProc] . procStatus == ProcInfo::STAT_WAITING) {
	  (*s) << kProc;
	  if(kProc < pInfo -> procData . size() - 1) (*s) << ",";
	}
      }
      break;
    }
    case LIFO_SCHED: 
    case ROUND_ROBIN_SCHED: {
      for(unsigned int kProc = 0; kProc < waitQueue . size(); ++kProc) {
	(*s) << waitQueue[kProc] ;
	if(kProc < waitQueue . size() - 1) (*s) << ",";
      }
    }
    case MULTILEVEL_FEEDBACK_SCHED: {
      for(unsigned int kQueue = 0 ; kQueue < mLevelQueue . size(); ++kQueue) {
	(*s) << kQueue << "(" ;
	for(unsigned int kProc = 0; kProc < mLevelQueue[kQueue] . size(); ++kProc) {
	  (*s) << mLevelQueue[kQueue][kProc] ;
	  if(kProc < mLevelQueue[kQueue] . size() - 1) (*s) << ",";
	}
	(*s) << ")";
	if(kQueue < mLevelQueue . size() - 1) (*s) << " ";
      }
      break;
    }
    default:{
      (*s) << "ERROR Unknown policy " << thePolicy << " in enQueueProc().\n";
      break;
    }
    }
    (*s) << ") ";
  }

  // la fonction "principale" de l'ordonnanceur

  int Scheduler::electAProc(int *allotedTime) {
    int chosenProc (ProcInfo::invalidProcPid);
    *allotedTime = allotedBaseTime;
    switch(thePolicy) {
    case LIFO_SCHED: {  // on prend par derriere, le dernier entre      
      displayQueue(&cerr);
      // TRAVAIL FAIT : - - - - - - - 
      // s'il y a des elements dans waitQueue
         // on prend le dernier de waitQueue dans 'chosenProc'
         // et onelimine ce dernier depuis waitQueue
      if(waitQueue . size()) {
	  chosenProc = waitQueue . back();
	  waitQueue . pop_back();
          // - - - - - 
	  pInfo -> procData[chosenProc] . computeWaitDuration();
      } // finsi
      return chosenProc;
    }
    default: {
      cerr << "ERROR Unknown policy " << thePolicy << " in electAProc().\n";
      break;
    }
    }
    return ProcInfo::invalidProcPid;
    
    }
  
  
}
