/* ordonnanceur-simulateur seulement avec LIFO */



#include <iostream> 
#include <fstream> 
#include <string>
#include <vector>
#include <deque> 
#include <sstream>
#include <signal.h>
#include <sys/types.h>  // pid_t
#include <sys/wait.h>
#include <unistd.h>     // getpid(),
#include <math.h>
#include <stdlib.h>
#include "ProcSched.h"
#include "CExc.h"
#include "nsSysteme.h"       // Fork()

using namespace nsSysteme;   // Fork()
using namespace ProcSched ;
using namespace std;

int main (int argc, char *argv[]) {
 try {

  if(argc != 3) {
    throw CExc("main()",string("Usage : ") +  argv[0] + 
		      " <prog name list file name> lifo\n");
  }
  const Scheduler::SchedPolicy policy(Scheduler::str2Policy(argv[2]));
  if(policy == Scheduler::INVALID_SCHED || policy != Scheduler::LIFO_SCHED) {
    cerr << "ERROR Invalid policy. Usage " << argv[0] << " <prog name list file name> lifo\n";
    return 2;
  }


  srand(getpid());
  time_t t;
  time(&t);
  double idleTime (0);
  int allotedTime (0);

  procInfo = new ProcInfo(argv[1]);
  procInfo -> displayProcInfo(&cerr);
  Scheduler scheduler(policy, procInfo);
  procInfo -> scheduler = &scheduler;
  scheduler . enQueueAllProc();

  // TRAVAIL FAIT : - - - - -
  // declarez et preparez le masque blockSigChld
  // pour pouvoir bloquer et debloquer plus tard SIGCHLD 
  sigset_t blockSigChld;      
  sigemptyset(&blockSigChld); 
  sigaddset  (&blockSigChld, SIGCHLD); 
  // - - - - - - - - - - - - 
  Signal(SIGCHLD, doIOTermAck);
  unsigned int userCPUTime (0);
  while(procInfo -> outstandingProcCount) {    
    // TRAVAIL FAIT : - - - 
    // bloquez SIGCHLD avec Sigprocmask()
    Sigprocmask(SIG_BLOCK, &blockSigChld, 0); 
    // affichez un petit compte rendu (Total "outstanding", etc.)
    cerr << "-- Total " 
	 << procInfo -> outstandingProcCount 
	 <<  " procs, now electing one among " ;
    // appelez la methode electAProc() et recuperez newProc2Run
    const int newProc2Run = scheduler . electAProc(&allotedTime);
    if(newProc2Run == ProcInfo::invalidProcPid) {
      cerr << ", got nobody.------------------\n";
      idleTime += allotedTime;
      // debloquez SIGCHLD avec Sigprocmask()
      Sigprocmask(SIG_UNBLOCK, &blockSigChld, 0);
      // - - - - - - - - - - - - 
      sleep(allotedTime);
      continue;
    }
    cerr << "got # " << newProc2Run << " " 
	 << procInfo -> getProcName(newProc2Run) << " -----------------\n";
    int workTime (procInfo -> updateProcData(newProc2Run, ProcInfo::ADVANCE_PROC, allotedTime));
    procInfo -> displayProcInfo(&cerr);
    Sigprocmask(SIG_UNBLOCK, &blockSigChld, 0);
    if(workTime > 0) { 
      userCPUTime += workTime;
      Sigprocmask(SIG_BLOCK, &blockSigChld, 0);
      procInfo -> updateProcData(newProc2Run, ProcInfo::CPU_TERM);
      Sigprocmask(SIG_UNBLOCK, &blockSigChld, 0);
    }
  }
  time(&t);
  cerr << "====================================\n";
  procInfo -> computeAverageWait();
  cout << "# " << scheduler . policy2str()  << " completed, user CPU: " 
       << userCPUTime
       << "s, Idle:" << idleTime 
       << "s, AvWait:" << procInfo -> averageWaitDuration
       << "s\n";
  procInfo -> dumpProcInfoStat(&cout);
  delete procInfo;
  return 0;
 }//fin try
 catch (const CExc & Exc) {
     cerr <<Exc<< endl;
     return errno;
 }
 catch (const exception & Exc) {
     cerr << "Exception : " << Exc.what () << endl;
     return 1;
 }
 catch (...) {
     cerr << "Exception inconnue recue dans la fonction main()" 
          << endl;
     return 1;
 }

}
